/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package FunctionLayer;

import DataLayer.DataMapper;
import FunctionLayer.StanfordParser.SentimentAnalyzerTest;
import com.google.common.base.Stopwatch;
import com.google.common.collect.MapMaker;
import edu.stanford.nlp.ie.AbstractSequenceClassifier;
import edu.stanford.nlp.ie.crf.CRFClassifier;
import edu.stanford.nlp.ling.CoreLabel;
import edu.stanford.nlp.ling.HasWord;
import edu.stanford.nlp.ling.TaggedWord;
import edu.stanford.nlp.ling.Word;
import edu.stanford.nlp.parser.lexparser.LexicalizedParser;
import edu.stanford.nlp.parser.shiftreduce.ShiftReduceParser;
import edu.stanford.nlp.pipeline.StanfordCoreNLP;
import edu.stanford.nlp.process.DocumentPreprocessor;
import edu.stanford.nlp.tagger.maxent.MaxentTagger;
import edu.stanford.nlp.trees.GrammaticalStructureFactory;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.trees.TreebankLanguagePack;
import java.io.IOException;
import java.io.StringReader;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author install1
 */
public class MYSQLDatahandler {

    public static final long EXPIRE_TIME_IN_SECONDS = TimeUnit.SECONDS.convert(10, TimeUnit.MINUTES);
    public static final long EXPIRE_TIME_IN_SECONDS1 = TimeUnit.SECONDS.convert(10, TimeUnit.HOURS);
    public static MYSQLDatahandler instance = new MYSQLDatahandler();
    public static int semeticsUpdateCount;
    private volatile boolean refreshMatrixFromDB;
    private final ConcurrentMap<Integer, String> stringCache;
    private LinkedHashMap<String, LinkedHashMap<String, Double>> lHMSMX = new LinkedHashMap();
    private final Stopwatch stopwatch;
    private final Stopwatch stopwatch1;
    private static String modelPath = "edu/stanford/nlp/models/srparser/englishSR.ser.gz";
    private static String sentimentModel = "edu/stanford/nlp/models/sentiment/sentiment.ser.gz";
    private static String lexParserEnglishRNN = "edu/stanford/nlp/models/lexparser/englishRNN.ser.gz";
    private static String taggerPath = "edu/stanford/nlp/models/pos-tagger/english-left3words/english-left3words-distsim.tagger";
    private static String NERModel = "edu/stanford/nlp/models/ner/english.all.3class.distsim.crf.ser.gz";
    private static MaxentTagger tagger;
    private static ShiftReduceParser model;
    private static String[] options = {"-maxLength", "100"};
    private static Properties props = new Properties();
    private static Properties propsSentiment = new Properties();
    private static GrammaticalStructureFactory gsf;
    private static LexicalizedParser lp;
    private static TreebankLanguagePack tlp;
    private static AbstractSequenceClassifier<CoreLabel> classifier;
    private ExecutorService executor;
    private static StanfordCoreNLP pipeline;
    private static StanfordCoreNLP pipelineSentiment;

    public static AbstractSequenceClassifier<CoreLabel> getClassifier() {
        return classifier;
    }

    public static void setClassifier(AbstractSequenceClassifier<CoreLabel> classifier) {
        MYSQLDatahandler.classifier = classifier;
    }

    public void instantiateExecutor() {
        this.executor = new ForkJoinPool(Runtime.getRuntime().availableProcessors(),
                ForkJoinPool.defaultForkJoinWorkerThreadFactory,
                null, true);
    }

    public MYSQLDatahandler() {
        this.stopwatch = Stopwatch.createUnstarted();
        this.stopwatch1 = Stopwatch.createStarted();
        this.stringCache = new MapMaker().concurrencyLevel(2).makeMap();
    }

    public static void shiftReduceParserInitiate() {
        try {
            classifier = CRFClassifier.getClassifierNoExceptions(NERModel);
        } catch (ClassCastException ex) {
            Logger.getLogger(MYSQLDatahandler.class.getName()).log(Level.SEVERE, null, ex);
        }
        model = ShiftReduceParser.loadModel(modelPath, options);
        tagger = new MaxentTagger(taggerPath);
        lp = LexicalizedParser.loadModel(lexParserEnglishRNN, options);
        tlp = lp.getOp().langpack();
        gsf = tlp.grammaticalStructureFactory();
        // set up pipeline properties
        props.setProperty("annotators", "tokenize,ssplit,pos,lemma,parse");
        props.setProperty("parse.model", modelPath);
        props.setProperty("parse.maxlen", "100");
        props.setProperty("tokenize.maxlen", "100");
        props.setProperty("ssplit.maxlen", "100");
        props.setProperty("lemma.maxlen", "100");
        props.setProperty("parse.binaryTrees", "true");
        propsSentiment.setProperty("annotators", "tokenize, ssplit, parse, sentiment");
        propsSentiment.setProperty("parse.model", lexParserEnglishRNN);
        propsSentiment.setProperty("sentiment.model", sentimentModel);
        propsSentiment.setProperty("sentiment.maxlen", "100");
        propsSentiment.setProperty("parse.maxlen", "100");
        propsSentiment.setProperty("tokenize.maxlen", "100");
        propsSentiment.setProperty("ssplit.maxlen", "100");
        // set up pipeline
        pipeline = new StanfordCoreNLP(props);
        pipelineSentiment = new StanfordCoreNLP(propsSentiment);
    }

    public static GrammaticalStructureFactory getGsf() {
        return gsf;
    }

    public static StanfordCoreNLP getPipeline() {
        return pipeline;
    }

    public static StanfordCoreNLP getPipelineSentiment() {
        return pipelineSentiment;
    }

    public static MaxentTagger getTagger() {
        return tagger;
    }

    public static ShiftReduceParser getModel() {
        return model;
    }

    private Map<Integer, String> getCache() throws SQLException, IOException, CustomError {
        List<String> strlist;
        strlist = DataMapper.getAllStrings();
        LinkedHashMap<Integer, String> LHM = new LinkedHashMap();
        int i = 0;
        for (String str : strlist) {
            LHM.put(i, str);
            i++;
        }
        return LHM;
    }

    public int getlHMSMXSize() {
        return lHMSMX.size();
    }

    public int getstringCacheSize() {
        return stringCache.size();
    }

    public void initiateMYSQL() throws SQLException, IOException {
        try {
            DataMapper.createTables();
            stringCache.putAll(getCache());
            lHMSMX = DataMapper.getAllRelationScores();
        } catch (CustomError ex) {
            Logger.getLogger(MYSQLDatahandler.class
                    .getName()).log(Level.SEVERE, null, ex);
        }
    }

    public synchronized void checkIfUpdateMatrixes() {
        refreshMatrixFromDB = false;
        int updateBadgesInteger = 900;
        int counter = 0;
        if (stopwatch1.elapsed(TimeUnit.SECONDS) >= EXPIRE_TIME_IN_SECONDS1) {
            refreshMatrixFromDB = true;
            lHMSMX = DataMapper.getAllRelationScores();
            stopwatch1.reset();
        }
        //requiring atleast 10 entries ensures no issues in case of empty stringcache
        if (stringCache.values().size() > 10 && !refreshMatrixFromDB) {
            ConcurrentMap<Integer, String> stringCachelocal = stringCache;
            int selectUpdate = -1;
            LinkedHashMap<String, LinkedHashMap<String, Double>> LHMSMXLocal = lHMSMX;
            int ij2 = 0;
            for (String str : stringCachelocal.values()) {
                if (!LHMSMXLocal.keySet().contains(str)) {
                    selectUpdate = ij2;
                    LinkedHashMap<String, Double> newLHM = new LinkedHashMap();
                    lHMSMX.put(str, newLHM);
                    break;
                }
                ij2++;
            }
            if (selectUpdate == -1) {
                ij2 = 0;
                int valueSize = stringCachelocal.size();
                for (LinkedHashMap<String, Double> values : LHMSMXLocal.values()) {
                    int lhmSize = values.size();
                    if (lhmSize < valueSize) {
                        selectUpdate = ij2;
                    }
                    ij2++;
                }
            }
            semeticsUpdateCount = selectUpdate;
            int beginindex = semeticsUpdateCount;
            System.out.println("beginindex: " + beginindex + "\n");
            ConcurrentMap<Integer, String> strIndexNavigator = new MapMaker().concurrencyLevel(2).makeMap();
            String get = stringCachelocal.get(beginindex);
            strIndexNavigator.put(strIndexNavigator.size() + 1, get);
            ConcurrentMap<Integer, String> strIndexAll = stringCachelocal;
            ConcurrentMap<Integer, String> randomIndexesToUpdate = new MapMaker().concurrencyLevel(2).makeMap();
            int indexes = updateBadgesInteger;
            if (indexes >= strIndexAll.size()) {
                indexes = strIndexAll.size() - 1;
            }
            int beginindexes = new Random().nextInt((strIndexAll.size()) - indexes);
            int ij1 = 0;
            while (beginindexes + ij1 < beginindexes + indexes) {
                String get1 = strIndexAll.get(beginindexes + ij1);
                randomIndexesToUpdate.put(ij1, get1);
                ij1++;
            }
            ConcurrentMap<Integer, SimilarityMatrix> matrixUpdateList = new MapMaker().concurrencyLevel(2).makeMap();
            ConcurrentMap<Integer, Future<SimilarityMatrix>> futures = new MapMaker().concurrencyLevel(2).makeMap();
            strIndexNavigator.values().forEach((str) -> {
                randomIndexesToUpdate.values().stream().filter((str1) -> (!str.equals(str1))).forEachOrdered((str1) -> {
                    boolean present = false;
                    if ((LHMSMXLocal.containsKey(str)) && LHMSMXLocal.get(str).containsKey(str1)
                            || (LHMSMXLocal.containsKey(str1)) && LHMSMXLocal.get(str).containsKey(str)) {
                        present = true;
                    }
                    if (!present) {
                        SimilarityMatrix SMX = new SimilarityMatrix(str, str1);
                        Callable<SimilarityMatrix> worker = new SentimentAnalyzerTest(str, str1, SMX);
                        futures.put(futures.size() + 1, executor.submit(worker));
                    }
                });
            });
            System.out.println("finished worker assignment, futures size: " + futures.size() + "\n");
            for (Future<SimilarityMatrix> future : futures.values()) {
                System.out.println("counter: " + counter + "\n");
                counter++;
                try {
                    SimilarityMatrix SMX = future.get(20, TimeUnit.SECONDS);
                    LinkedHashMap<String, Double> getFuture = lHMSMX.getOrDefault(SMX.getPrimaryString(), null);
                    if (getFuture == null) {
                        getFuture = new LinkedHashMap();
                    }
                    getFuture.put(SMX.getSecondaryString(), SMX.getDistance());
                    lHMSMX.put(SMX.getPrimaryString(), getFuture);
                    matrixUpdateList.put(matrixUpdateList.size() + 1, SMX);
                } catch (InterruptedException | ExecutionException | TimeoutException ex) {
                    Logger.getLogger(MYSQLDatahandler.class.getName()).log(Level.SEVERE, null, ex);
                }
            }
            try {
                if (!matrixUpdateList.isEmpty()) {
                    DataMapper.insertSementicMatrixes(matrixUpdateList);
                    System.out.println("finished datamapper semetic insert");
                }
            } catch (CustomError ex) {
                Logger.getLogger(MYSQLDatahandler.class
                        .getName()).log(Level.SEVERE, null, ex);
            }
        }
    }

    public synchronized void checkIfUpdateStrings() throws CustomError {
        if (stopwatch.elapsed(TimeUnit.SECONDS) >= EXPIRE_TIME_IN_SECONDS || !stopwatch.isRunning()) {
            new Thread(() -> {
                ConcurrentMap<Integer, String> str = MessageResponseHandler.getStr();
                str = cutContent(str);
                str = filterContent(str);
                str = removeSlacks(str);
                try {
                    DataMapper.InsertMYSQLStrings(str);

                } catch (CustomError ex) {
                    Logger.getLogger(MYSQLDatahandler.class
                            .getName()).log(Level.SEVERE, null, ex);
                }
                MessageResponseHandler.setStr(new MapMaker().concurrencyLevel(2).makeMap());
                int j = stringCache.size() + 1;
                for (String str1 : str.values()) {
                    stringCache.put(j, str1);
                    j++;
                }
            }).start();
            if (!stopwatch.isRunning()) {
                stopwatch.start();
            } else {
                stopwatch.reset();
            }
        }
    }

    public synchronized String getResponseMsg(String str) throws CustomError {
        str = str.trim();
        if (str.startsWith("<@")) {
            str = str.substring(str.indexOf("> ") + 2);
        }
        final LinkedHashMap<String, LinkedHashMap<String, Double>> LHMSMXLocal = lHMSMX;
        ConcurrentMap<Integer, String> strArrs = stringCache;
        double Score = -10000;
        SimilarityMatrix SMXreturn = new SimilarityMatrix("", "");
        System.out.println("pre mostSimilarSTR \n");
        String mostSimilarSTR = mostSimilar(str, strArrs);
        if (mostSimilarSTR != null && !mostSimilarSTR.isEmpty()) {
            System.out.println("mostSimilarSTR; " + mostSimilarSTR + "\n");
            LinkedHashMap<String, Double> orDefault = LHMSMXLocal.getOrDefault(mostSimilarSTR, null);
            if (orDefault != null) {
                for (Entry<String, Double> entrySet : orDefault.entrySet()) {
                    double smxDistance = entrySet.getValue();
                    if (smxDistance > Score) {
                        Score = smxDistance;
                        SMXreturn = new SimilarityMatrix(mostSimilarSTR, entrySet.getKey(), smxDistance);
                    }
                }
            }
            for (Entry<String, LinkedHashMap<String, Double>> values1 : LHMSMXLocal.entrySet()) {
                LinkedHashMap<String, Double> value = values1.getValue();
                for (Entry<String, Double> keystr : value.entrySet()) {
                    if (keystr.getKey().equals(mostSimilarSTR)) {
                        double smxDistance = keystr.getValue();
                        if (smxDistance > Score) {
                            Score = smxDistance;
                            SMXreturn = new SimilarityMatrix(values1.getKey(), keystr.getKey(), smxDistance);
                        }
                    }
                }
            }
            if (!SMXreturn.getPrimaryString().isEmpty()) {
                if (SMXreturn.getPrimaryString().equals(mostSimilarSTR)) {
                    return SMXreturn.getSecondaryString();
                } else {
                    return SMXreturn.getPrimaryString();
                }
            }
        }
        System.out.println("none within 8 range");
        ConcurrentMap<Integer, String> strCache = stringCache;
        ConcurrentMap<Integer, Future<SimilarityMatrix>> futureslocal = new MapMaker().concurrencyLevel(2).makeMap();
        for (String str1 : strCache.values()) {
            if (!str.equals(str1)) {
                Callable<SimilarityMatrix> worker = new SentimentAnalyzerTest(str, str1, new SimilarityMatrix(str, str1));
                futureslocal.put(futureslocal.size() + 1, executor.submit(worker));
            }
        }
        int index = 0;
        for (Future<SimilarityMatrix> future : futureslocal.values()) {
            try {
                SimilarityMatrix SMX = future.get((long) 20, TimeUnit.SECONDS);
                double distance = SMX.getDistance();
                System.out.println("index: " + index + "\nfutures size: " + futureslocal.values().size() + "\nScore: " + SMX.getDistance() + "\nSecondary: "
                        + SMX.getSecondaryString() + "\nPrimary: " + SMX.getPrimaryString() + "\n");
                if (distance > Score) {
                    Score = distance;
                    SMXreturn = SMX;
                }
            } catch (InterruptedException | ExecutionException | TimeoutException ex) {
                System.out.println("ex: " + ex.getMessage() + "\n");
            }
            index++;
        }
        System.out.println("Reached end: secondary: " + SMXreturn.getSecondaryString() + "\nPrimarY: " + SMXreturn.getPrimaryString()
                + "\nScore: " + SMXreturn.getDistance());
        return SMXreturn.getSecondaryString();
    }

    public String mostSimilar(String toBeCompared, ConcurrentMap<Integer, String> concurrentStrings) {
        int minDistance = 8;
        String similar = "";
        ConcurrentMap<Integer, Future<DistanceObject>> futures = new MapMaker().concurrencyLevel(2).makeMap();
        concurrentStrings.values().stream().map((str) -> new LevenshteinDistance(str, toBeCompared, new DistanceObject())).forEachOrdered((worker) -> {
            futures.put(futures.size() + 1, executor.submit(worker));
        });
        for (Future<DistanceObject> future : futures.values()) {
            try {
                DistanceObject d = future.get();
                int distance = d.getDistance();
                System.out.println("distance: " + distance + "\n");
                if (distance < minDistance) {
                    minDistance = distance;
                    similar = d.getSentence();
                }
            } catch (NullPointerException | InterruptedException | ExecutionException ex) {
                System.out.println("failed future\n");
            }
        }
        return similar;
    }

    public static ConcurrentMap<Integer, String> cutContent(ConcurrentMap<Integer, String> str) {
        ConcurrentMap<Integer, String> returnlist = new MapMaker().concurrencyLevel(2).makeMap();
        for (String str1 : str.values()) {
            int iend = str1.indexOf("content: ");
            if (iend != -1) {
                String trs = str1.substring(iend + 9);
                returnlist.put(returnlist.size() + 1, trs.substring(0, trs.length() - 1));
            }
        }
        return returnlist;
    }

    public static ConcurrentMap<Integer, String> filterContent(ConcurrentMap<Integer, String> str) {
        ConcurrentMap<Integer, String> strlistreturn = new MapMaker().concurrencyLevel(2).makeMap();
        for (String str1 : str.values()) {
            if (str1.isEmpty() || str1.length() < 3) {
                continue;
            }
            str1 = str1.trim();
            if (str1.contains("PM*")) {
                str1 = str1.substring(str1.indexOf("PM*") + 5);
            }
            if (str1.contains("AM*")) {
                str1 = str1.substring(str1.indexOf("AM*") + 5);
            }
            for (Character c : str1.toCharArray()) {
                if (c == '?' || c == '°') {
                    str1 = str1.replace("?", " <:wlenny:514861023002624001> ");
                    str1 = str1.replace("°", " <:wlenny:514861023002624001> ");
                }
            }
            if (str1.contains("(Counter-Terrorist)")) {
                str1 = str1.replace("(Counter-Terrorist)", " ");
            }
            if (str1.contains("(Terrorist)")) {
                str1 = str1.replace("(Terrorist)", " ");
            }
            if (str1.contains("(Spectator)")) {
                str1 = str1.replace("(Spectator)", " ");
            }
            if (str1.contains("*DEAD*")) {
                str1 = str1.replace("*DEAD*", " ");
            }
            if (str1.contains("{red}")) {
                str1 = str1.replace("{red}", " ");
            }
            if (str1.contains("{orange}")) {
                str1 = str1.replace("{orange}", " ");
            }
            if (str1.contains("{yellow}")) {
                str1 = str1.replace("{yellow}", " ");
            }
            if (str1.contains("{green}")) {
                str1 = str1.replace("{green}", " ");
            }
            if (str1.contains("{lightblue}")) {
                str1 = str1.replace("{lightblue}", " ");
            }
            if (str1.contains("{blue}")) {
                str1 = str1.replace("{blue}", " ");
            }
            if (str1.contains("{purple}")) {
                str1 = str1.replace("{purple}", " ");
            }
            if (str1.contains("{white}")) {
                str1 = str1.replace("{white}", " ");
            }
            if (str1.contains("{fullblue}")) {
                str1 = str1.replace("{fullblue}", " ");
            }
            if (str1.contains("{cyan}")) {
                str1 = str1.replace("{cyan}", " ");
            }
            if (str1.contains("{lime}")) {
                str1 = str1.replace("{lime}", " ");
            }
            if (str1.contains("{deeppink}")) {
                str1 = str1.replace("{deeppink}", " ");
            }
            if (str1.contains("{slategray}")) {
                str1 = str1.replace("{slategray}", " ");
            }
            if (str1.contains("{dodgerblue}")) {
                str1 = str1.replace("{dodgerblue}", " ");
            }
            if (str1.contains("{black}")) {
                str1 = str1.replace("{black}", " ");
            }
            if (str1.contains("{orangered}")) {
                str1 = str1.replace("{orangered}", " ");
            }
            str1 = str1.trim();
            if (str1.length() > 2 && (!str1.startsWith("!"))) {
                strlistreturn.put(strlistreturn.size() + 1, str1);
            }
        }
        return strlistreturn;
    }

    private ConcurrentMap<Integer, String> removeSlacks(ConcurrentMap<Integer, String> str) {
        ShiftReduceParser model = getModel();
        MaxentTagger tagger = getTagger();
        List<TaggedWord> taggedWords;
        ConcurrentMap<Integer, String> strreturn = new MapMaker().concurrencyLevel(2).makeMap();
        for (String str1 : str.values()) {
            int counter = 0;
            ConcurrentMap<Integer, String> TGWList = new MapMaker().concurrencyLevel(2).makeMap();
            DocumentPreprocessor tokenizer = new DocumentPreprocessor(new StringReader(str1));
            for (List<HasWord> sentence : tokenizer) {
                List<TaggedWord> tagged1 = tagger.tagSentence(sentence);
                Tree tree = model.apply(tagged1);
                taggedWords = tree.taggedYield();
                for (TaggedWord TGW : taggedWords) {
                    if (!TGWList.values().contains(TGW.tag()) && !TGW.tag().equals(":") && !TGW.word().equals(TGW.tag())) {
                        TGWList.put(TGWList.size() + 1, TGW.tag());
                        counter++;
                    }
                    if (counter > 3) {
                        int addCounter = 0;
                        ConcurrentMap<Integer, Word> wordList = new MapMaker().concurrencyLevel(2).makeMap();
                        for (Word lab : tree.yieldWords()) {
                            if (lab != null && lab.word() != null) {
                                //System.out.println("lab: " + lab + " \n");
                                if (!wordList.values().contains(lab) && lab.value() != null && !lab.value().equals(":")) {
                                    wordList.put(wordList.size() + 1, lab);
                                    addCounter++;
                                }
                            }
                        }
                        if (addCounter > 3) {
                            addCounter = 0;
                            ConcurrentMap<Integer, HasWord> HWlist = new MapMaker().concurrencyLevel(2).makeMap();
                            for (HasWord HW : tree.yieldHasWord()) {
                                if (HW != null && HW.word() != null && !HWlist.values().contains(HW)) {
                                    addCounter++;
                                    HWlist.put(HWlist.size() + 1, HW);
                                }
                            }
                            if (addCounter > 3) {
                                boolean tooclosematch = false;
                                Collection<String> values = stringCache.values();
                                for (String strVals : values) {
                                    LevenshteinDistance leven = new LevenshteinDistance(strVals, str1);
                                    double Distance = leven.computeLevenshteinDistance();
                                    int maxpermittedDistance = 2;
                                    if (Distance < maxpermittedDistance) {
                                        tooclosematch = true;
                                        break;
                                    }
                                }
                                if (!tooclosematch) {
                                    strreturn.put(strreturn.size() + 1, str1);
                                }
                            }
                        }
                        break;
                    }
                }
                if (counter > 3) {
                    break;
                }
            }
        }
        return strreturn;
    }
}
